#!/usr/bin/env python
import base64
import logging
from pathlib import Path

import exifread
import pandas as pd

from puddl.db.alchemy import App

log = logging.getLogger()

img_dir = Path('/home/felix/Seafile/s7-camera/My Photos/Camera/')
imgs = list(img_dir.glob('*.jpg'))
# imgs = list(img_dir.glob('20200831_160821.jpg'))
log.info(f'processing {len(imgs)} images')


def img2dict(path: Path):
    result = {
        '_path': str(path),
        '_name': path.name
    }
    with path.open('rb') as f:
        result.update(exifread.process_file(f))
    if 'JPEGThumbnail' in result:
        payload = base64.b64encode(result['JPEGThumbnail']).decode('utf-8')
        img = 'data:image/jpeg;base64,' + payload
    else:
        img = 'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQ' \
              'AAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII='
    result['thumb'] = img
    return result


def _exifread2pg_dict(d):
    result = {}
    for k, v in d.items():
        if isinstance(v, exifread.classes.IfdTag):
            result[k] = v.printable
        else:
            result[k] = v
    return result


def iter_imgs(imgs):
    for img in imgs:
        d = img2dict(img)
        yield _exifread2pg_dict(d)


app = App('exif')
log.info(app.engine)

df = pd.DataFrame(iter_imgs(imgs))
app.df_dump(df, 's7')



import re
from typing import TextIO


class ParseError(Exception):
    pass


RE_TIMESTAMP = re.compile(r'^#(\d{5}\d+)', re.MULTILINE | re.DOTALL)


def iter_bash_history(fd: TextIO):
    """
    :param fd: file descriptor of a bash history file with HISTTIMEFORMAT='%F__%H-%M-%S '
    :return: (timestamp as int, command_line) tuples
    """
    data = fd.read()
    flat_ts2cmd = RE_TIMESTAMP.split(data)
    # e.g. ["ignore this command", "1585254366", "\nfoo\n", "1585254441", "\nbar\n"]
    # skip to the first timestamp
    while not flat_ts2cmd[0].isnumeric():
        flat_ts2cmd.pop(0)
    # e.g. ["1585254366", "\nfoo\n", "1585254441", "\nbar\n"]
    for ts, command_line in zip(flat_ts2cmd[:-1:2], flat_ts2cmd[1::2]):
        # i.e. zip(["1585254366", "1585254441"], ["\nfoo\n", "\nbar\n"])
        yield int(ts), command_line.strip()



import datetime
import logging
import re

import click
from flask import Flask, request, make_response
from sqlalchemy.exc import SQLAlchemyError

from puddl import get_config
from puddl.db.alchemy import get_engine
from .models import Base, schema, Thing

log = logging.getLogger()


@click.group()
def main():
    pass


@main.command()
def run():
    conf = get_config(app='mampf')
    schema.bind(get_engine(conf))
    Base.metadata.create_all(schema.engine)

    app = Flask(__name__)

    @app.route('/')
    def index():
        response = make_response(f'POST stuff to {request.base_url}<scope>\n')
        response.headers['Content-Type'] = 'text/plain'
        return response

    @app.route('/<scope>', methods=['POST'])
    def mjam(scope):
        session = schema.Session()
        t = Thing(
            scope=scope,
            x=dict(request.form),
        )
        try:
            session.add(t)
            session.commit()
            response = make_response()  # no news is good news
            response.headers['Content-Type'] = 'text/plain'
            return response
        except SQLAlchemyError:
            session.rollback()
            response = make_response('NOK\n', 500)
            response.headers['Content-Type'] = 'text/plain'
            return response

    app.run()


def _parse_since(x):
    RE_SINCE = re.compile(
        r'((?P<days>\d+?)d)?'
        r'((?P<hours>\d+?)h)?'
        r'((?P<minutes>\d+?)m)?'
        r'((?P<seconds>\d+?)s)?')
    m = RE_SINCE.match(x)
    td_kw = {}
    for k, v in m.groupdict().items():
        if v is None:
            continue
        else:
            td_kw[k] = float(v)
    return datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(**td_kw)


@main.command()
@click.option('--limit', default=100)
@click.option(
    '--since',
    help="Only return logs newer than a relative duration like 5s, 2m, or 3h."
)
def ls(limit, since):
    conf = get_config(app='mampf')
    schema.bind(get_engine(conf))
    session = schema.Session()
    things = session.query(Thing) \
        .order_by(Thing.dt.desc())
    # time filter
    if since is not None:
        then = _parse_since(since)
        things = things.filter(Thing.dt > then)
    # limit
    things = things.limit(limit)
    for thing in reversed(things.all()):
        print(thing)



import logging
import subprocess

import click

from puddl import get_config
from puddl.conf import DBConfig

log = logging.getLogger(__name__)


@click.group()
def db():
    pass


@db.command()
@click.option('--app')
def health(app):
    conf = DBConfig(app)

    import psycopg2
    # consumes PG* env vars
    connection = psycopg2.connect(conf.url)
    with connection.cursor() as cursor:
        cursor.execute('SELECT 1')
    log.info('database available')


@db.command()
@click.option('--app')
def shell(app):
    conf = DBConfig(app)
    conf.PGAPPNAME = 'puddl db shell'
    # want to see good error handling for this kind of thing?
    # https://github.com/pallets/click/blob/19fdc8509a1b946c088512e0e5e388a8d012f0ce/src/click/_termui_impl.py#L472-L487
    subprocess.Popen('psql', env=conf.to_dict()).wait()


@db.command()
@click.option('--app')
def url(app):
    """
    print DB URL

    Useful for things like this:

        from sqlalchemy import create_engine
        database_url = 'postgresql://puddl:aey1Ki4oaZohseWod2ri@localhost:13370/puddl'
        engine = create_engine(database_url, echo=False)
        df.to_sql('bp', engine)
    """
    conf = DBConfig(app)
    print(conf.url)


@db.command()
@click.option('--app')
def env(app):
    """
    Prints the database's environment.
    WARNING: This dumps your password. Use it like this:

        source <(puddl db env)
    """
    conf = DBConfig(app)
    for k, v in conf.to_dict().items():
        print(f'export {k}={v}')


@db.command()
def sessions():
    """
    Lists active sessions.
    """
    conf = get_config(app='db sessions')
    query = """SELECT pid AS process_id,
           usename AS username,
           datname AS database_name,
           client_addr AS client_address,
           application_name,
           backend_start,
           state,
           state_change
    FROM pg_stat_activity;"""
    click.echo(
        subprocess.check_output(['psql', '-c', query], encoding='utf-8', env=conf.db_env))



import logging
import socket

import uuid

from django.db import models

log = logging.getLogger(__name__)


class Datum(models.Model):

    date = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True
        ordering = ['-last_modified']


class DeviceManager(models.Manager):
    def me(self):
        device, created = self.get_or_create(name=socket.gethostname())
        if created:
            log.info(f'created device {device}')
        else:
            log.debug(f'using device {device}')
        return device


class Device(models.Model):
    objects = DeviceManager()
    date = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)

    name = models.CharField(max_length=40, unique=True)

    class Meta:
        db_table = 'device'
        ordering = ['-last_modified']

    def __str__(self):
        return self.name



import logging

import click

import puddl
from puddl.apps import App
from puddl.conf import get_config

log = logging.getLogger(__name__)
LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG']


class AppsGroup(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        self._apps = [App(x) for x in get_config().apps]
        self._name2app = {
            x.cmd_name: x for x in self._apps
        }
        super().__init__(name, commands, **attrs)

    def list_commands(self, ctx):
        core_commands = set(super().list_commands(ctx))
        app_names = set(x for x in self._name2app)
        return sorted(core_commands | app_names)

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._name2app:
            app = self._name2app[cmd_name]
            return app.get_command()
        return super().get_command(ctx, cmd_name)


@click.group(cls=AppsGroup)
@click.option('-l', '--log-level',
              type=click.Choice(LOG_LEVELS, case_sensitive=False),
              default='WARNING')
@click.option('-d', '--debug', is_flag=True,
              help='sets log level to DEBUG. ignores "--log-level"')
@click.option('--loggers', default='', help='comma-separated logger names')
@click.version_option(version=puddl.__version__)
def root(log_level, debug, loggers):
    if debug:
        log_level = 'DEBUG'
    logging.basicConfig(level=log_level)
    if loggers:
        for name in loggers.split(','):
            logging.getLogger(name).setLevel(log_level)
    pass



# Generated by Django 3.0.4 on 2020-03-27 00:03

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('file', '0002_auto_20200315_1105'),
    ]

    operations = [
        migrations.CreateModel(
            name='BashHistory',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('dt', models.DateTimeField()),
                ('command', models.TextField()),
                ('file', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='file.File')),
            ],
            options={
                'db_table': 'bash_history',
            },
        ),
    ]



import logging

from puddl.cli.app import app
from puddl.cli.config import config
from puddl.cli.db import db
from puddl.cli.root import root

log = logging.getLogger(__name__)

# every module defines a group. add them here.
root.add_command(config)
root.add_command(app)
root.add_command(db)


def main():
    # this function is referenced in `setup.py`
    root(auto_envvar_prefix='PUDDL')


if __name__ == '__main__':
    main()



import json
import logging

import click
import pandas as pd

from puddl.db.alchemy import App

log = logging.getLogger()


@click.group()
def main():
    pass


def _iter_lines2dicts(lines):
    for line in lines:
        data = json.loads(line)
        yield data


@main.command()
@click.argument('schema')
@click.argument('table')
def load(schema, table):
    app = App(schema)
    df = pd.DataFrame(_iter_lines2dicts(click.get_text_stream('stdin')))
    app.df_dump(df, table, index=False)
    log.info(f'SELECT * FROM {table}')



import socket
from pathlib import Path

from filetype import filetype


def default_device_name():
    return socket.gethostname()


def guess_filetype(path: Path):
    x = filetype.guess(str(path))
    if x is not None:
        return {
            "extension": x.extension,
            "mime": x.mime
        }


def get_stat(path: Path):
    s = path.stat()
    attrs = [a for a in dir(s) if a.startswith('st_')]
    return {a: getattr(s, a) for a in attrs}



#!/usr/bin/env python
import logging
import os
import runpy
from pathlib import Path

import click
import sqlalchemy

from puddl.db import get_alchemy_session
from puddl.db.alchemy import get_engine

log = logging.getLogger()


class CodiMDError(Exception):
    pass


@click.group()
def main():
    pass


def get_remotes(ctx, args, incomplete):
    from puddl.felix.codimd.models import Remote
    session = get_alchemy_session()
    return session.query(Remote.name).all()


@main.command()
@click.argument('remote_name', autocompletion=get_remotes, envvar='CODIMD_NAME')
def index(remote_name):
    os.environ['CODIMD_NAME'] = remote_name
    # Run the "script"; it should not be imported, but be run from an interpreter.
    # We could create a subprocess, but choose to use this process' interpreter instead.
    runpy.run_module('puddl.felix.codimd.index')


@main.command()
def ls():
    import pandas as pd
    engine = get_engine()
    df = pd.read_sql("SELECT * FROM codimd.cli_ls", engine, index_col='id')
    print(df)


@main.group()
def db():
    pass


@db.command()
def create():
    from puddl.felix.codimd.models import Base
    logging.getLogger('sqlalchemy').setLevel(log.level)
    engine = get_engine()
    engine.execute('CREATE SCHEMA IF NOT EXISTS codimd')
    Base.metadata.create_all(engine)
    views_path = Path(__file__).parent / 'views.sql'
    with views_path.open(encoding='utf-8') as f:
        engine.execute(sqlalchemy.text(f.read()))


@db.command()
@click.option('-f', '--force', is_flag=True)
def drop(force):
    q = 'DROP SCHEMA codimd CASCADE'
    if not force:
        click.confirm(f'Will run "{q}"; are you sure?', abort=True)
    logging.getLogger('sqlalchemy').setLevel(log.level)
    engine = get_engine()
    engine.execute(q)



#!/usr/bin/env python

import json

from puddl.db.alchemy import App

app = App('exif')

data = []
rows = app.engine.execute('SELECT * FROM markers')

for row in rows:
    d = dict(zip(rows.keys(), row))
    data.append(d)

jsdata = json.dumps(data)
print(f"window.exif_data = {jsdata}")



import logging

from django.db import models
from django.utils import timezone

from puddl.contrib.file.models import File
from .lib import iter_bash_history
from puddl.models import Datum

log = logging.getLogger(__name__)


class BashHistoryManager(models.Manager):

    def from_file(self, file: File):
        with file.path.open() as f:
            rows = iter_bash_history(f)
            instances = []
            for ts, command_line in rows:
                instances.append(self.model(
                    file=file,
                    dt=timezone.datetime.fromtimestamp(ts),
                    command=command_line,
                ))
            return self.bulk_create(instances)


class BashHistory(Datum):
    objects = BashHistoryManager()

    file = models.ForeignKey(File, on_delete=models.CASCADE)
    dt = models.DateTimeField()
    command = models.TextField()

    class Meta:
        db_table = 'bash_history'

    def __str__(self):
        return f'{self.file}@{self.dt}'



import json
import logging
import os
from collections import UserDict
from pathlib import Path
from urllib.parse import urlencode

from puddl.apps import App

log = logging.getLogger(__name__)


class ConfigError(Exception):
    pass


class Config(UserDict):
    PUDDLRC = Path.home() / '.puddlrc'

    DB_DEFAULTS = {
        'PGUSER': os.environ.get('PGUSER'),
        'PGPASSWORD': os.environ.get('PGPASSWORD'),
        'PGHOST': os.environ.get('PGHOST'),
        'PGPORT': os.environ.get('PGPORT'),
        'PGDATABASE': os.environ.get('PGDATABASE'),
        'PGAPPNAME': os.environ.get('PGAPPNAME', 'puddl'),
    }

    @classmethod
    def get_defaults(cls):
        defaults = cls.DB_DEFAULTS.copy()
        defaults['apps'] = []
        return defaults

    def write(self):
        x = json.dumps(self.data, indent=2, sort_keys=True)
        with self.PUDDLRC.open('w') as f:
            f.write(x)

    @property
    def apps(self):
        return self['apps']

    @property
    def db_url(self) -> str:
        db_params = {
            'application_name': self['PGAPPNAME']
        }
        return 'postgresql://{PGUSER}:{PGPASSWORD}@{PGHOST}:{PGPORT}' \
               '/{PGDATABASE}?{params}'.format(**self, params=urlencode(db_params))

    @property
    def db_env(self):
        return {k: v for k, v in self.items() if k in self.DB_DEFAULTS.keys()}

    def add_app(self, name):
        p = App(name)
        p.validate()
        self['apps'].append(p.pkg_name)

    def remove_app(self, name):
        ps = self['apps']
        try:
            del ps[ps.index(name)]
        except ValueError:
            log.warning(f'app {name} was not installed')

    def get_app_from_module_name(self, module: str):
        """
        Given `puddl.felix.codimd.cli` return `puddl.felix.codimd` if `puddl.felix.codimd`
        is installed.
        """
        for installed_app_name in self['apps']:
            if module.startswith(installed_app_name):
                return App(installed_app_name)
        return None


class DBConfig:
    """
    Abstracts a subset of the libpq environment variables defined by
    https://www.postgresql.org/docs/current/libpq-envars.html

    >>> puddl_root_db_config = DBConfig()
    >>> puddl_root_db_config.url
    'postgresql://puddl:aey1Ki4oaZohseWod2ri@localhost:13370/puddl?application_name=puddl'

    Given an app name, user and password are set accordingly

    >>> app_db_config = DBConfig('foo')
    >>> app_db_config.url
    'postgresql://foo:foo-aey1Ki4oaZohseWod2ri@localhost:13370/puddl?application_name=foo'
    """
    __slots__ = ['PGUSER', 'PGPASSWORD', 'PGHOST', 'PGPORT', 'PGDATABASE', 'PGAPPNAME']

    def __init__(self, name=None):
        env_vars_missing = []
        for key in self.__slots__:
            try:
                setattr(self, key, os.environ[key])
            except KeyError as e:
                key = e.args[0]
                env_vars_missing.append(key)

        conf_vars_missing = []
        if env_vars_missing:
            conf = get_config()
            for key in env_vars_missing:
                try:
                    value = conf[key]
                    setattr(self, key, value)
                except KeyError:
                    conf_vars_missing.append(key)

        if conf_vars_missing:
            raise ConfigError(f'Missing variables: {env_vars_missing}. You may set them '
                              f'in the environment or by initializing ~/.puddlrc')
        if name is not None:
            self.PGUSER = name
            self.PGPASSWORD = '-'.join((name, self.PGPASSWORD))
            self.PGAPPNAME = name

    def to_dict(self):
        return {k: getattr(self, k) for k in self.__slots__}

    @property
    def url(self) -> str:
        db_params = {
            'application_name': self.PGAPPNAME
        }
        params = urlencode(db_params)
        return f'postgresql://{self.PGUSER}:{self.PGPASSWORD}' \
               f'@{self.PGHOST}:{self.PGPORT}' \
               f'/{self.PGDATABASE}?{params}'

    def __str__(self):
        return self.PGUSER

    def __repr__(self):
        return f"DBConfig('{self.PGUSER}')"


def get_config(app=None):
    try:
        with Config.PUDDLRC.open() as f:
            data = json.load(f)
    except FileNotFoundError:
        data = Config.get_defaults()
    except json.decoder.JSONDecodeError as e:
        raise ConfigError(
            f'{Config.PUDDLRC}:{e.lineno} column {e.colno} char {e.pos}: {e.msg}'
        )
    cfg = Config(**data)
    # app must be one of the installed apps
    if app is None:
        app = 'puddl'

    data['PGAPPNAME'] = app
    return cfg



from pathlib import Path

import click


@click.command(name='bash_history')
@click.argument('path', type=click.Path(exists=True),
                default=Path('~/.bash_history').expanduser())
def main(path):
    from puddl.contrib.file.models import Device, File
    file = File.objects.upsert(
        device=Device.objects.me(),
        path=path
    )
    from .models import BashHistory
    BashHistory.objects.from_file(file)



import inspect
import subprocess
from itertools import chain

import sqlalchemy
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import CreateColumn
from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.types import Text

from puddl import get_config
from puddl.conf import DBConfig


def _get_caller_app():
    # We need to go down the stack two times:
    #
    # - we are called by something (our caller)
    # - someone else called our caller - that's what we want
    frm = inspect.stack()[2]
    mod_name = inspect.getmodule(frm[0]).__name__
    conf = get_config()
    return conf.get_app_from_module_name(mod_name)


def get_engine(conf=None, **kwargs):
    """
    Convenience function to get an engine based on puddl config.
    """

    if conf is None:
        conf = get_config()

    return sqlalchemy.create_engine(conf.db_url, **kwargs)


def get_alchemy_session(conf=None, **kwargs):
    """
    Convenience. Don't know if this is wise. :>
    """
    engine = get_engine(conf, **kwargs)
    session_cls = sessionmaker(bind=engine)
    return session_cls()


# https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#postgresql-10-identity-columns
@compiles(CreateColumn, 'postgresql')
def use_identity(element, compiler, **kw):
    text = compiler.visit_create_column(element, **kw)
    text = text.replace("SERIAL", "INT GENERATED BY DEFAULT AS IDENTITY")
    return text


class Upserter:
    def __init__(self, session, model):
        self.session = session
        self.model = model
        inspector = Inspector.from_engine(session.bind)
        _xs = inspector.get_unique_constraints(self.model.__tablename__,
                                               schema=self.model.__table__.schema)
        list_of_lists = [x['column_names'] for x in _xs]
        self.unique_columns = set(chain.from_iterable(list_of_lists))

    def upsert(self, **data):
        # read this carefully:
        # https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#sqlalchemy.dialects.postgresql.dml.Insert.on_conflict_do_update
        columns = set(data.keys())
        update_columns = columns - self.unique_columns
        insert_stmt = insert(self.model).values(**data)
        excluded_mapping = {  # LHS: column name, RHS: Column instance
            col: getattr(insert_stmt.excluded, col)
            for col in update_columns
        }
        do_update_stmt = insert_stmt.on_conflict_do_update(
            index_elements=self.unique_columns,
            set_=excluded_mapping,
        )
        self.session.execute(do_update_stmt)


class Schema:
    """
    This has two usages: in models for declarative_base and interactively (CLI,
    notebook, ...).

    In models, you care about the `declarative_base` and some meta info like `name`.
    You do not have an engine, so you cannot supply a bind. In that case, just don't::

        >>> schema = Schema('foo')

    When you use the schema interactively, you want to operate against the database.
    You must have a bind, so provide it. The schema gets created automatically and
    switches the connection to use it:

        >>> engine = ...
        >>> schema = Schema('foo', bind=engine)

    You can then use the engine like this:

        >>> schema.engine.execute('SELECT 1').first()
        (1,)
    """
    engine = None
    Session = None

    def __init__(self, schema_name, bind=None):
        self.name = schema_name
        if bind is not None:
            self.bind(bind)

    def declarative_base(self):
        class Base:
            __table_args__ = {"schema": self.name}

        return declarative_base(cls=Base)

    def bind(self, engine):
        self.engine = engine
        self.engine.execute(f'CREATE SCHEMA IF NOT EXISTS {self.name}')
        self.engine.execute(f'SET search_path TO {self.name}')
        self.Session = sessionmaker(bind=self.engine)


# https://docs.sqlalchemy.org/en/13/core/functions.html
# noinspection PyPep8Naming
class puddl_upsert_role(GenericFunction):
    type = Text
    package = 'puddl'
    identifier = 'upsert_role'


class App:
    """
    >>> app = App('sunrise')

    This creates a Postgres ROLE and SCHEMA with the name "sunrise".

    Please instantiate this class once per process, because it uses SQLAlchemy's engine,
    thus similar rules apply. "[...] the Engine is most efficient when created just once
    at the module level of an application, not per-object or per-function call"
    -- https://docs.sqlalchemy.org/en/13/core/connections.html#basic-usage

    An app holds DB configuration and an engine:

    >>> app.db_config
    DBConfig('sunrise')
    >>> app.engine
    Engine(postgresql://sunrise:***@localhost:13370/puddl?application_name=sunrise)
    """
    name: str
    db_config: DBConfig
    engine: Engine

    def __init__(self, name):
        self.name = name

        # get global credentials (to upsert the app itself)
        puddl_root_db_config = DBConfig()
        my_password = f'{self.name}-{puddl_root_db_config.PGPASSWORD}'
        root_engine = sqlalchemy.create_engine(puddl_root_db_config.url)
        session_cls = sessionmaker(root_engine)
        session = session_cls()
        try:
            session.execute(f"CREATE SCHEMA IF NOT EXISTS {name}")
            session.execute(sqlalchemy.func.puddl.upsert_role(name, my_password))
            session.execute(f"GRANT ALL ON SCHEMA {name} TO {name}")
            # puddl may write to any scope
            session.execute(f"GRANT ALL ON SCHEMA {name} TO puddl")
            session.commit()
        except ProgrammingError:
            session.rollback()
            raise
        self.db_config = DBConfig(self.name)
        self.engine = sqlalchemy.create_engine(self.db_config.url)

    def has_table(self, table_name):
        return self.engine.dialect.has_table(self.engine, table_name)

    def df_dump(self, df, table_name, index=True):
        """
        Dumps a DataFrame to table_name.
        WARNING! This truncates the table if it exists.
        """
        if self.has_table(table_name):
            self.engine.execute(f'TRUNCATE TABLE {table_name}')
        if index:
            # follow SQL conventions for the index
            df.index += 1
        return df.to_sql(
            table_name, self.engine, self.name, if_exists='replace',
            index=index
        )

    def psql(self):
        db_env = self.db_config.to_dict()
        subprocess.Popen('psql', env=db_env).wait()



from importlib import import_module


class AppError(Exception):
    pass


class App:
    """
    Must be a package that contains a click.Command object called "main".
    """

    def __init__(self, pkg_name: str):
        self.pkg_name = pkg_name
        # use the package name as cmd_name
        self.cmd_name = self.pkg_name.split('.')[-1]
        self.pkg = import_module(pkg_name)

    @property
    def version(self):
        return self.pkg.__version__

    def __str__(self):
        return f'{self.pkg_name}=={self.version}'

    def get_command(self):
        cmd = self.pkg.main
        # The app's command might have define a different name by mistake, e.g.
        # @click.command(name='something-completely-different')
        # Let's force it into compliance.
        cmd.name = self.cmd_name
        return cmd

    def validate(self):
        """
        Currently just checks if the app can be imported.
        """
        try:
            self.get_command()
        except ImportError:
            raise AppError(f'Not found: {self}')



from pathlib import Path

from sqlalchemy import DateTime, func, Column, Integer, String, types, JSON, ForeignKey, \
    UniqueConstraint
from sqlalchemy.orm import relationship

from puddl.db.alchemy import Schema
from .lib import default_device_name

schema = Schema('file')
Base = schema.declarative_base()


# maybe class LastModifiedMixin
# onupdate=func.now()

class DateMixin:
    # server_default=func.sysdate()
    # https://docs.sqlalchemy.org/en/13/core/defaults.html#server-invoked-ddl-explicit-default-expressions
    date = Column(DateTime(timezone=True), default=func.now())


class Device(Base, DateMixin):
    __tablename__ = 'device'
    id = Column(Integer, primary_key=True)
    name = Column(String(40), unique=True, default=default_device_name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'Device(id={self.id}, name={self.name})'


# https://docs.sqlalchemy.org/en/13/core/custom_types.html
# noinspection PyAbstractClass


class PathType(types.TypeDecorator):
    impl = types.Text

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return Path(value)


class File(Base, DateMixin):
    __tablename__ = 'file'
    id = Column(Integer, primary_key=True)
    device_id = Column(Integer, ForeignKey(Device.id))

    path = Column(PathType, nullable=False)
    filetype = Column(JSON)
    stat = Column(JSON, nullable=False)

    device = relationship(Device)
    UniqueConstraint(device_id, path)


def get_default_device(session):
    """
    Gets the local device based on hostname.
    Creates it, if it does not exist yet.
    """
    name = default_device_name()
    device = session.query(Device).filter(Device.name == name).first()
    if not device:
        device = Device(name=name)
        session.add(device)
    return device



from pathlib import Path
from unittest import TestCase

from puddl.felix.bash_history.lib import iter_bash_history


class Test(TestCase):
    HERE = Path(__file__).parent
    MULTILINE = """\
cat <<EOF > /tmp/hello
world
EOF\
"""

    def test_hist2dict(self):
        path = self.HERE / 'example.bash_history'
        with path.open() as f:
            actual = list(iter_bash_history(f))
        expected = [
            (1585254366, self.MULTILINE),
            (1585254441, 'tail ~/.bash_history'),
            (1585254727, "# this is a comment\necho and it's valid."),
        ]
        self.assertEqual(expected, actual)



"""
Indexes a CodiMD remote given as os environment variable CODIMD_NAME.

Can be run standalone like this::

    CODIMD_NAME=foo python index.py
"""
import logging
import os

import pandas as pd
import requests

from puddl import get_config
from puddl.db import get_alchemy_session
from puddl.db.alchemy import Upserter
from puddl.felix.codimd import CodiMDError
from puddl.felix.codimd.models import Remote, Note, History

log = logging.getLogger(__name__)
conf = get_config(app='codimd')

remote_name = os.environ['CODIMD_NAME']

db_session = get_alchemy_session(conf)
s = requests.session()

# I'd like to use "?" here, but providing create_engine with paramstyle='qmark'
# does not work. :(
# https://docs.sqlalchemy.org/en/13/core/engines.html#sqlalchemy.create_engine.params.paramstyle
remote = db_session.query(Remote) \
    .filter(Remote.name == remote_name) \
    .one()


def raise_on_alert_in_dom(response: requests.Response):
    """
    CodiMD seems not to know how to return non-200 HTTP status codes.
    So we look for "alert" in the DOM. *sigh*
    Could use beautiful soup or lxml for that, but meh.
    """
    o = response.text.find('alert')
    if o >= 0:
        raise CodiMDError(response.text[o:o + 200])


s.post(f'{remote.url}/login',
       data={'email': remote.email, 'password': remote.password})

r = s.get(f'{remote.url}/history')
raise_on_alert_in_dom(r)
records = r.json()['history']
# [{'id': 'pycharm', 'text': 'pycharm', 'time': 1582037284079, 'tags': []}, ...]
log.debug(records)
df = pd.DataFrame(records)
# db.bind.execute('DELETE FROM codimd.history')
db_session.bind.execute('TRUNCATE codimd.history RESTART IDENTITY CASCADE')
df.to_sql(
    'history', db_session.bind,
    schema='codimd', if_exists='append', index=False,
    dtype={k: v.type for k, v in History.__table__.columns.items()}
)
# want to get this in a notebook?
# df = pd.read_sql('SELECT * FROM codimd.history', engine, index_col='index')

ups = Upserter(db_session, Note)

for note_id in df['id']:
    log.info(f'fetching {note_id}')
    r = s.get(f'{remote.url}/{note_id}/download')

    ups.upsert(
        remote_id=remote.id,
        note_id=note_id,
        text=r.text,
    )
db_session.commit()



import click

from puddl import get_config
from puddl.apps import App


@click.group()
def app():
    pass


@app.command('add')
@click.argument('package_path')
def app_add(package_path):
    conf = get_config()
    conf.add_app(package_path)
    conf.write()


@app.command('ls')
@click.option('-v', '--verbose', is_flag=True)
def app_ls(verbose):
    conf = get_config()
    apps = sorted(conf['apps'])
    for app_name in apps:
        msg = app_name
        if verbose:
            msg = str(App(app_name))
        click.echo(msg)


@app.command('rm')
@click.argument('package_path')
def app_remove(package_path):
    conf = get_config()
    conf.remove_app(package_path)
    conf.write()



from sqlalchemy import Column, Integer, Text, String, ForeignKey, UniqueConstraint, \
    BigInteger
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship

from puddl.db.alchemy import Schema

schema = Schema('codimd')

Base = schema.declarative_base()


# https://docs.sqlalchemy.org/en/13/orm/tutorial.html
class Remote(Base):
    __tablename__ = 'remotes'

    id = Column(Integer, primary_key=True, unique=True)
    name = Column(String(20), unique=True)
    url = Column(Text, unique=True)
    email = Column(Text)
    password = Column(Text)

    def __str__(self):
        return f'Remote({self.id} {self.name} {self.url} {self.email})'


class History(Base):
    """
    As given by CODIMD_URL/history
    """
    __tablename__ = 'history'

    id = Column(Text, primary_key=True)
    tags = Column(postgresql.ARRAY(Text))
    text = Column(Text)
    time = Column(BigInteger)

    # TODO remote_id


class Note(Base):
    __tablename__ = 'notes'

    id = Column(Integer, primary_key=True)
    remote_id = Column(Integer, ForeignKey(Remote.id))
    note_id = Column(Text)
    text = Column(Text)

    remote = relationship('Remote')

    UniqueConstraint(remote_id, note_id)



#!/usr/bin/env python
from setuptools import setup, find_packages, find_namespace_packages

setup(
    packages=find_packages() + find_namespace_packages(include=['puddl.*']),
    install_requires=[
        'Click',
        'flask',
        'ffmpeg-python',
        'filetype',
        'pandas',
        'psycopg2-binary',
        'requests',
        'sqlalchemy',
        'tinytag',
    ],
    entry_points='''
        [console_scripts]
        puddl=puddl.cli:main
    ''',
)



import logging
from pathlib import Path

import click

from puddl.conf import Config, get_config

log = logging.getLogger(__name__)


@click.group()
def config():
    pass


@config.command()
def init():
    data = {}
    defaults = Config.get_defaults()
    data.update(defaults)

    dot_env = Path('.env')
    try:
        with dot_env.open() as f:
            contents = f.read()
        for line in contents.strip().split('\n'):
            k, v = line.strip().split('=')
            if k in defaults:
                log.info(f'using {k} from .env')
                data[k] = v
    except FileNotFoundError:
        pass

    conf = Config(data)
    conf.write()


@config.command()
def show():
    conf = get_config()
    print(conf.data)



import logging
from pathlib import Path

import click
from sqlalchemy.sql import ddl

from puddl.db import get_alchemy_session
from puddl.felix.file.db import get_default_device, File
from puddl.db.alchemy import Upserter, get_engine
from puddl.felix.file.lib import guess_filetype, get_stat

log = logging.getLogger(__name__)


@click.group(name='file')
def main():
    pass


def _iter_stdin():
    with open(0, 'rb') as f:
        for bytewurst in f.read().split(b'\n'):
            try:
                line = bytewurst.decode('utf-8')
                if line != '':
                    yield Path(line)
            except UnicodeDecodeError:
                log.warning(f'could not decode line {line}')


@main.command()
@click.argument('path', type=click.Path(), nargs=-1)
def index(path):
    """
    Index a file.

    Consume STDIN if no arguments given. This is useful for pipelines like this:

        find . -type f -name '*.py' | puddl file index
    """
    # Using click's metavar makes things unnecessarily complicated. We want to keep Backus
    # Naur in the generated help output.
    paths = path
    del path

    def _iter_paths():
        if not paths:
            yield from _iter_stdin()
        else:
            for x in paths:
                yield Path(x)

    session = get_alchemy_session()
    device = get_default_device(session)
    ups = Upserter(session, File)
    for path in _iter_paths():
        log.debug(path)
        ups.upsert(
            path=path,
            device_id=device.id,
            filetype=guess_filetype(path),
            stat=get_stat(path)
        )
    session.commit()


@main.command()
@click.option('--limit', default=100)
@click.option('--all', is_flag=True)
def ls(limit, all):
    session = get_alchemy_session()
    query = session.query(File)
    if not all:
        query = query.limit(limit)
    for f in query:
        print(f.path)


@main.command('query')
@click.argument('query_string', type=click.STRING)
def q(query_string):
    """
    query by "path contains"
    """
    session = get_alchemy_session()
    for x in session.query(File).filter(File.path.contains(query_string)):
        print(x.path)


@main.group()
def db():
    pass


@db.command()
def create():
    from puddl.felix.file.db import Base
    logging.getLogger('sqlalchemy').setLevel(log.level)
    engine = get_engine()
    engine.execute(ddl.CreateSchema('file'))
    Base.metadata.create_all(engine)


@db.command()
@click.option('-f', '--force', is_flag=True)
def drop(force):
    statement = ddl.DropSchema('file', cascade=True)
    if not force:
        click.confirm(f'Will run "{statement}"; are you sure?', abort=True)
    logging.getLogger('sqlalchemy').setLevel(log.level)
    engine = get_engine()
    engine.execute(statement)



from sqlalchemy import DateTime, func, Column, Integer, JSON, Text

from puddl.db.alchemy import Schema

schema = Schema('mampf')
Base = schema.declarative_base()


class Thing(Base):
    __tablename__ = 'thing'
    id = Column(Integer, primary_key=True, unique=True)
    dt = Column(DateTime(timezone=True), default=func.now())
    scope = Column(Text, nullable=False)
    x = Column(JSON, nullable=False)

    def __str__(self):
        local_dt = self.dt.astimezone()
        return f'id={self.id} dt={local_dt} scope={self.scope} x={self.x}'
